@@ -1,4 +1,4 @@
-# -*- coding: euc-jp -*-
+# -*- coding: utf-8 -*-
 require "tkcanvas"
 
 if defined?($hscale_demo) && $hscale_deom
@@ -12,21 +12,22 @@ $hscale_demo = TkToplevel.new {|w|
 }
 positionWindow($hscale_demo)
 
+base_frame = TkFrame.new($hscale_demo).pack(:fill=>:both, :expand=>true)
 
-msg = TkLabel.new($hscale_demo) {
+msg = TkLabel.new(base_frame) {
   font $font
   wraplength '3.5i'
   justify 'left'
-  text "ˤ1Ĥȿʿʥ뤬ɽƤޤ\
-ǥޥܥ1򥯥åޤϥɥå\
-ĹѤ뤳ȤǤޤ"
+  text "下には矢印が1つと水平なスケールが表示されています。\
+スケール上でマウスボタン1をクリック、またはドラッグすると\
+矢印の長さを変えることができます。"
 }
 msg.pack('side'=>'top')
 
-TkFrame.new($hscale_demo) {|frame|
+TkFrame.new(base_frame) {|frame|
   TkButton.new(frame) {
-    #text 'λ'
-    text 'Ĥ'
+    #text '了解'
+    text '閉じる'
     command proc {
       tmppath = $hscale_demo
       $hscale_demo = nil
@@ -35,12 +36,22 @@ TkFrame.new($hscale_demo) {|frame|
   }.pack('side'=>'left', 'expand'=>'yes')
 
   TkButton.new(frame) {
-    text 'ɻ'
+    text 'コード参照'
     command proc { showCode 'hscale' }
   }.pack('side'=>'left', 'expand'=>'yes')
 }.pack('side'=>'bottom', 'fill'=>'x', 'pady'=>'2m')
 
-TkFrame.new($hscale_demo) {|frame|
+def setWidth(w, width)
+  width = width + 21
+  x2 = width - 30
+  if x2 < 21
+    x2 = 21
+  end
+  w.coords 'poly',20,15,20,35,x2,35,x2,45,width,25,x2,5,x2,15,20,15
+  w.coords 'line',20,15,20,35,x2,35,x2,45,width,25,x2,5,x2,15,20,15
+end
+
+TkFrame.new(base_frame) {|frame|
   canvas = TkCanvas.new(frame) {|c|
     width 50
     height 50
@@ -65,14 +76,3 @@ TkFrame.new($hscale_demo) {|frame|
   }.pack('side'=>'bottom', 'expand'=>'yes', 'anchor'=>'n')
   scale.set 75
 }.pack('side'=>'top', 'fill'=>'x')
-
-
-def setWidth(w, width)
-  width = width + 21
-  x2 = width - 30
-  if x2 < 21
-    x2 = 21
-  end
-  w.coords 'poly',20,15,20,35,x2,35,x2,45,width,25,x2,5,x2,15,20,15
-  w.coords 'line',20,15,20,35,x2,35,x2,45,width,25,x2,5,x2,15,20,15
-end
